/*
	PROforma, PROgs FOnt & Raster MAnager

	PROforma colour control
	PFGrayShade
	PFColourRGB
	PFColourCMYK
	PFPaperGrayShade
	PFPaperColourRGB
	PFPaperColourCMYK

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 September 1995
	21.10.1995 the origin of a fastpattern is now adjusted according to the
			   PageOrigin to make it match when scrolling etc
			   ColourAdjust now always called, and test for use of Black or
			   White driver only done after adjusting the colour. This allows
			   sharing a bitmap driver between (mono) screen and printer.
	20.12.1995 made colour conversion routines explicit (also used by pictures)
	07.08.1996 introduction of colour cache
	03.12.1996 patternmask now used when building fastpattern
*/

#include <array.h>
#include "gstate.h"

#define catch(x) do { if (err=(x)) return err; } while (0)

/* --- colour conversion routines --- */

Error Gray2Gray(pt *from, pt *to)
{ *to=*from; return ERR_OK; }
Error Gray2RGB(pt *from, ColourRGB *to)
{
	to->red=to->green=to->blue=pt_hundred-*from;
	return ERR_OK;
}
Error Gray2CMYK(pt *from, ColourCMYK *to)
{
	to->cyan=to->magenta=to->yellow=0;
	to->black=*from;
	return ERR_OK;
}
Error RGB2Gray(ColourRGB *from, pt *to)
{
	*to=pt_hundred-( fixmul(from->red  ,double2pt(.30)) +
					 fixmul(from->green,double2pt(.59)) +
					 fixmul(from->blue ,double2pt(.11)) );
	return ERR_OK;
}
Error RGB2RGB(ColourRGB *from, ColourRGB *to)
{ *to=*from; return ERR_OK; }
Error RGB2CMYK(ColourRGB *from, ColourCMYK *to)
{
	pt ceiling;
#define LIMIT(x) if (x<0) x=0; else if (x>pt_hundred) x=pt_hundred
	to->black  =pt_hundred-max(from->red,max(from->green,from->blue));
	LIMIT(to->black);
	ceiling=pt_hundred-to->black;
	to->cyan   =ceiling-from->red;;
	LIMIT(to->cyan);
	to->magenta=ceiling-from->green;
	LIMIT(to->magenta);
	to->yellow =ceiling-from->blue;
	LIMIT(to->yellow);
	return ERR_OK;
#undef LIMIT
}
Error CMYK2Gray(ColourCMYK *from, pt *to)
{
	*to=from->black+fixmul(from->cyan	,double2pt(.30)) +
					fixmul(from->magenta,double2pt(.59)) +
					fixmul(from->yellow ,double2pt(.11));
	if (*to<0) *to=0; else if (*to>pt_hundred) *to=pt_hundred;
	return ERR_OK;
}
Error CMYK2RGB(ColourCMYK *from, ColourRGB *to)
{
	to->red  =pt_hundred-min(pt_hundred,from->cyan	 +from->black);
	to->green=pt_hundred-min(pt_hundred,from->magenta+from->black);
	to->blue =pt_hundred-min(pt_hundred,from->yellow +from->black);
	return ERR_OK;
}
Error CMYK2CMYK(ColourCMYK *from, ColourCMYK *to)
{ *to=*from; return ERR_OK; }

/* --- FastPattern handling --- */

Error PatternOrigin(Gstate gstate)
{
	short xorg, yorg;
	PBbox *box=&gstate->raster->patternbox;

	xorg=abs(pt2short(gstate->Origin.x)-gstate->raster->clip.xorg) % box->xsiz;
	yorg=abs(pt2short(gstate->Origin.y)-gstate->raster->clip.yorg) % box->ysiz;

	if ((xorg!=box->xorg && yorg!=box->yorg) || !gstate->raster->bitdraw)
		return ERR_OK;

	return gstate->raster->bitdraw->ChangeOrigin(gstate->raster,xorg,yorg);
}

typedef struct { short count; int colour; } ctable;

static int Ccolour(const void *n1,const void *n2)
{
	return ((ctable *)n1)->count-((ctable *)n2)->count;
}

Error buildFastPattern(Gstate gstate, ColourCMYK *colour, CMYKRouts *routs)
{
	struct _Bitmap bitmap;
	BITMdriver *bit=gstate->bitmap;
	unsigned int i, j;
	int devcol;
	ctable colours[64];
	int colcount=0;
	ColourCMYK col, delta;

	bitmap.base=gstate->raster->pattern;
	bitmap.linc=gstate->raster->pattlinc;
	bitmap.pinc=gstate->raster->pattpinc;
	MEMCopy(sizeof(ColourCMYK),colour,&col);

	/* determine which colours occur and their multiplicity */

	for (i=64 ; i ; i--)
	{
		routs->ColourNearest(gstate,&col,&devcol);
		routs->ColourDelta(gstate,&col,&delta);
		col.cyan   =colour->cyan   +delta.cyan;
		col.magenta=colour->magenta+delta.magenta;
		col.yellow =colour->yellow +delta.yellow;
		col.black  =colour->black  +delta.black;
		for (j=0 ; j<colcount && colours[j].colour!=devcol ; j++);
		if (j==colcount) colours[j].count=0, colours[j].colour=devcol, colcount++;
		colours[j].count++;
	}
	ARRAYSort(colours,colcount,sizeof(ctable),Ccolour);

	bit->CursorSet(&bitmap,0,0);
	for (i=gstate->raster->patternbox.ysiz ; i ; i--)
	{
		for (j=gstate->raster->patternbox.xsiz ; j ; j--)
		{
			short curr=gstate->patternmask[i%8][j%8], count, k;
			if (curr>63) curr=63;	/* quick sanity check */
			for (k=count=0 ; (count+=colours[k].count)<curr ; k++);
			bit->PixelSetNextX(&bitmap,colours[k].colour);
		}
	}
	gstate->raster->patternbox.xorg=gstate->raster->patternbox.yorg=0;
	return PatternOrigin(gstate);
}

Error GetFastPattern(Gstate gstate, ColourCMYK *colour, CMYKRouts *routs)
{
	Error err;
	ColourCache *start, *curr;

	/* start by searching for the colour in the cache */
	start=curr=gstate->ColCache;
	for (curr=start->next ; curr!=start ; curr=curr->next)
	{
		if (MEMSame(sizeof(ColourCMYK),&curr->colour,colour))
		{
			/* unlink colour */
			curr->next->prev=curr->prev;
			curr->prev->next=curr->next;
			/* relink at start */
			curr->next=start->next;
			start->next->prev=curr;
			start->next=curr;
			curr->prev=start;
			/* set as current fastpattern */
			gstate->raster->pattern=curr->pattern;
			MEMCopy(sizeof(PBbox),&curr->patternbox,&gstate->raster->patternbox);
			return PatternOrigin(gstate);
		}
	}
	/* oops, not found -> build colour in least recently used cache entry */
	curr=start->prev;
	gstate->raster->pattern=curr->pattern;
	/* and build it */
	catch( buildFastPattern(gstate,colour,routs) );
	MEMCopy(sizeof(PBbox),&gstate->raster->patternbox,&curr->patternbox);
	MEMCopy(sizeof(ColourCMYK),colour,&curr->colour);
	return ERR_OK;
}

/* --- routines to set the display colour and paper colour --- */

Error PFColourGray(Gstate gstate, pt shade)
{
	Error err=ERR_OK;
	ColourRGB rgb;
	ColourCMYK cmyk;

	/* action depends on native colour space for the device */
	switch(gstate->printer->ColourSpace)
	{
	case PF_COLOURSPACE_GRAYSHADE:
		gstate->drawcolourspace=PF_COLOURSPACE_GRAYSHADE;
		gstate->DrawColour.black=shade;

		gstate->printer->colour.Gray.ColourAdjust(gstate,&shade);
		if (shade==0)
			gstate->raster->bitdraw=&gstate->bitmap->White;
		else if (shade==pt_hundred)
			gstate->raster->bitdraw=&gstate->bitmap->Black;
		else
		{
			MEMClear(sizeof(ColourCMYK),&cmyk);
			cmyk.cyan=shade;
			
			/* build a FastPattern for this grayshade */
			gstate->raster->bitdraw=&gstate->bitmap->FastPattern;
			GetFastPattern(gstate,&cmyk,&gstate->printer->colour.CMYK);
		}
		break;
	case PF_COLOURSPACE_RGB:
		Gray2RGB(&shade,&rgb);
		err=PFColourRGB(gstate,&rgb);
		break;
	case PF_COLOURSPACE_CMY:
	case PF_COLOURSPACE_CMYK:
		Gray2CMYK(&shade,&cmyk);
		err=PFColourCMYK(gstate,&cmyk);
		break;
	default:
		err=ERR_ISYN;
		break;
	}
	return err;
}

static ColourRGB RGBWhite={pt_hundred,pt_hundred,pt_hundred};
static ColourRGB RGBBlack={0,0,0};

Error PFColourRGB(Gstate gstate, const ColourRGB *_colour)
{
	Error err=ERR_OK;
	ColourRGB colour;
	ColourCMYK cmyk;
	pt shade;
	colour=*_colour;

	/* action depends on native colour space for the device */
	switch(gstate->printer->ColourSpace)
	{
	case PF_COLOURSPACE_GRAYSHADE:
		RGB2Gray(&colour,&shade);
		err=PFColourGray(gstate,shade);
		break;
	case PF_COLOURSPACE_RGB:
		gstate->drawcolourspace=PF_COLOURSPACE_RGB;
		MEMCopy(sizeof(ColourRGB),&colour,&gstate->DrawColour);

		gstate->printer->colour.RGB.ColourAdjust(gstate,&colour);
		if (MEMSame(sizeof(ColourRGB),&colour,&RGBWhite))
			gstate->raster->bitdraw=&gstate->bitmap->White;
		else if (MEMSame(sizeof(ColourRGB),&colour,&RGBBlack))
			gstate->raster->bitdraw=&gstate->bitmap->Black;
		else
		{
			MEMCopy(sizeof(ColourRGB),&colour,&cmyk);
			cmyk.black=0;
			/* build a FastPattern for this colour */
			gstate->raster->bitdraw=&gstate->bitmap->FastPattern;
			GetFastPattern(gstate,&cmyk,&gstate->printer->colour.CMYK);
		}
		break;
	case PF_COLOURSPACE_CMY:
	case PF_COLOURSPACE_CMYK:
		RGB2CMYK(&colour,&cmyk);
		err=PFColourCMYK(gstate,&cmyk);
		break;
	default:
		err=ERR_ISYN;
		break;
	}
	return err;
}

static ColourCMYK CMYKWhite={0,0,0,0};
static ColourCMYK CMYKBlack={0,0,0,pt_hundred};

Error PFColourCMYK(Gstate gstate, const ColourCMYK *_colour)
{
	Error err=ERR_OK;
	ColourCMYK colour;
	ColourRGB rgb;
	pt shade;
	colour=*_colour;

	/* action depends on native colour space for the device */
	switch(gstate->printer->ColourSpace)
	{
	case PF_COLOURSPACE_GRAYSHADE:
		CMYK2Gray(&colour,&shade);
		err=PFColourGray(gstate, shade);
		break;
	case PF_COLOURSPACE_RGB:
		CMYK2RGB(&colour,&rgb);
		err=PFColourRGB(gstate,&rgb);
		break;
	case PF_COLOURSPACE_CMY:
	case PF_COLOURSPACE_CMYK:
		gstate->drawcolourspace=PF_COLOURSPACE_CMYK;
		MEMCopy(sizeof(ColourCMYK),&colour,&gstate->DrawColour);

		gstate->printer->colour.CMYK.ColourAdjust(gstate,&colour);
		if (MEMSame(sizeof(ColourCMYK),&colour,&CMYKWhite))
			gstate->raster->bitdraw=&gstate->bitmap->White;
		else if (MEMSame(sizeof(ColourCMYK),&colour,&CMYKBlack))
			gstate->raster->bitdraw=&gstate->bitmap->Black;
		else
		{
			/* build a FastPattern for this colour */
			gstate->raster->bitdraw=&gstate->bitmap->FastPattern;
			GetFastPattern(gstate,&colour,&gstate->printer->colour.CMYK);
		}
		break;
	default:
		err=ERR_ISYN;
		break;
	}
	return err;
}

Error PFPaperColourGray(Gstate gstate, pt shade)
{
	gstate->papercolourspace=PF_COLOURSPACE_GRAYSHADE;
	gstate->PaperColour.black=shade;
	return ERR_OK;
}

Error PFPaperColourRGB(Gstate gstate, const ColourRGB *colour)
{
	gstate->papercolourspace=PF_COLOURSPACE_RGB;
	return MEMCopy(sizeof(ColourRGB),colour,&gstate->PaperColour);
}

Error PFPaperColourCMYK(Gstate gstate, const ColourCMYK *colour)
{
	gstate->papercolourspace=PF_COLOURSPACE_CMYK;
	return MEMCopy(sizeof(ColourCMYK),colour,&gstate->PaperColour);
}
